"""Log line for symbol_please."""
import re

import dateutil.parser

LOG_REGEX = re.compile(r'^\[(?P<time>[^\]]+)\] (?P<message>.+)$')


class LogLine():
    """An object that represents a single log line."""

    def __init__(self, line, line_id):
        """Initialize the log line from its string representation."""
        match = LOG_REGEX.match(line)
        if match is None:
            raise ValueError(
                "Given line does not match expected log format: '{}'"
                .format(line))

        self.line_id = line_id
        self.raw_line = line
        self.raw_timestamp = match.group('time')
        self.timestamp = dateutil.parser.parse(self.raw_timestamp)
        self.message = match.group('message')

    def __repr__(self):
        return self.message



"""Spell data module for symbol_please."""
import json
import os

DATA_FILE = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), 'spell_data.json')

with open(DATA_FILE, 'r') as data_file:
    SPELL_DATA = json.load(data_file)



"""User interface module for symbol_please."""
from collections import OrderedDict
import datetime
import logging
import queue

import cairo
import gi

from symbol_please.parser import ParserThread
from symbol_please.parser.spell import SpellEventCast, SpellEventExpired

gi.require_version('Gtk', '3.0')
from gi.repository import GLib, Gtk, Gdk  # noqa: E402

_LOGGER = logging.getLogger(__name__)

REMOVE_IMAGE = Gtk.Image(stock=Gtk.STOCK_REMOVE)


def effect_remaining_time(effect):
    """Calculate remaining time for effect."""
    now = datetime.datetime.now()
    cast_time = effect.cast_time
    duration = effect.duration

    elapsed_time = (now - cast_time).total_seconds()
    return duration - elapsed_time


class OverlayWindow(Gtk.Window):
    """Main overlay window."""

    def __init__(self, profile_name, config):
        """Initialize the main window."""
        super().__init__(title="Symbol Please")

        self.connect("destroy", Gtk.main_quit)

        self.parser = ParserThread(profile_name, config)
        self.parser.start()

        self._event_queue = self.parser.client.add_event_listener_thread()
        self._active_effects = OrderedDict()
        self._effect_hboxes = {}
        self._effect_bars = {}
        self._effect_remove_btns = {}

        self.locked = False

        self.set_decorated(False)
        self.set_size_request(500, 0)
        self.set_type_hint(Gdk.WindowTypeHint.UTILITY)
        self.set_keep_above(True)

        self.set_events(Gdk.EventMask.BUTTON_PRESS_MASK)
        self.connect('button-press-event', self.on_button_press)

        screen = self.get_screen()
        visual = screen.get_rgba_visual()
        self.set_visual(visual)

        self.lock_button = Gtk.Button(label="Lock")
        self.lock_button.connect("clicked", self.on_lock_clicked)

        self.vbox = Gtk.VBox(spacing=5)
        self.vbox.set_spacing(5)

        tophbox = Gtk.HBox()
        tophbox.pack_start(self.lock_button, False, False, 5)

        quit_image = Gtk.Image(stock=Gtk.STOCK_QUIT)
        self.quit_button = Gtk.Button(image=quit_image)
        self.quit_button.connect("clicked", Gtk.main_quit)
        tophbox.pack_end(self.quit_button, False, False, 5)

        self.vbox.pack_start(tophbox, False, False, 0)

        self.contentvbox = Gtk.VBox()
        self.contentvbox.set_spacing(5)
        self.vbox.pack_start(self.contentvbox, True, True, 5)

        self.add(self.vbox)

        self.set_opacity(0.8)

        GLib.timeout_add(25, self.update)

    def on_button_press(self, widget, event):
        """On main window button press."""
        # Only handle left clicks
        if event.button == 1:
            if event.x <= 20:
                # If clicking the left side, begin a resize
                self.get_window().begin_resize_drag_for_device(
                    Gdk.WindowEdge.WEST, event.device, event.button,
                    event.x_root, event.y_root, event.time)
            elif self.get_size().width - event.x <= 20:
                # If clicking the right side, begin a resize
                self.get_window().begin_resize_drag_for_device(
                    Gdk.WindowEdge.EAST, event.device, event.button,
                    event.x_root, event.y_root, event.time)
            else:
                # Otherwise begin a move
                self.get_window().begin_move_drag_for_device(
                    event.device, event.button, event.x_root, event.y_root,
                    event.time)
            return True

    def on_lock_clicked(self, widget):
        """On lock button clicked."""
        self.locked = not self.locked
        # self.set_app_paintable(self.locked)
        if self.locked:
            self.lock_button.set_label("Unlock")
            self.set_opacity(0.5)
            self.lock_allocate_handler = self.lock_button.connect(
                "size-allocate", self.on_lock_size_allocate)
            self.quit_button.hide()
        else:
            self.lock_button.set_label("Lock")
            self.set_opacity(0.8)
            self.quit_button.show()

        for btn in self._effect_remove_btns.values():
            if self.locked:
                btn.hide()
            else:
                btn.show()

    def on_lock_size_allocate(self, widget, area):
        """On lock button clicked."""
        self.lock_button.disconnect(self.lock_allocate_handler)
        if self.locked:
            rectangle_int = cairo.RectangleInt(
                area.x, area.y, area.width, area.height)
            self.get_window().input_shape_combine_region(
                cairo.Region(rectangle_int), 0, 0)

    def update(self):
        """Run periodic updates on UI thread."""
        self.check_event_queue()
        self.update_effects()
        return True

    def update_effects(self):
        """Repack the effect progress bars."""
        for instance_id in self._active_effects.keys():
            name = self._active_effects[instance_id].spell['name']
            target = self._active_effects[instance_id].target

            remaining_time = effect_remaining_time(
                self._active_effects[instance_id])
            duration = self._active_effects[instance_id].duration

            self._effect_bars[instance_id].set_fraction(
                remaining_time / duration)

            if remaining_time >= 100:
                remaining_mins = int(remaining_time / 60)
                self._effect_bars[instance_id].set_text(
                    "{} - {} ({}m)".format(name, target, remaining_mins))
            elif remaining_time > 0:
                remaining_secs = int(remaining_time)
                self._effect_bars[instance_id].set_text(
                    "{} - {} ({}s)".format(name, target, remaining_secs))
            elif remaining_time <= -100:
                expired_mins = int(remaining_time * -1 / 60)
                self._effect_bars[instance_id].set_text(
                    "{} - {} (Expired {}m)".format(name, target, expired_mins))
            else:
                expired_secs = int(remaining_time * -1)
                self._effect_bars[instance_id].set_text(
                    "{} - {} (Expired {}s)".format(name, target, expired_secs))

    def redraw_effects(self):
        """Repack the effect progress bars."""
        self._active_effects = OrderedDict(sorted(
            self._active_effects.items(),
            key=lambda item: effect_remaining_time(item[1])))
        self.update_effects()

        for child in self.contentvbox:
            self.contentvbox.remove(child)

        for instance_id in self._active_effects.keys():
            self.contentvbox.pack_start(
                self._effect_hboxes[instance_id], False, False, 0)

            self._effect_hboxes[instance_id].show()
            self._effect_bars[instance_id].show()
            if not self.locked:
                self._effect_remove_btns[instance_id].show()
        width = self.get_size()[0]
        self.resize(width, 1)

    def check_event_queue(self):
        """Check the event queue for new events."""
        try:
            while True:
                event = self._event_queue.get_nowait()
                self.on_parser_event(event)
        except queue.Empty:
            pass

    def on_remove_click(self, instance_id):
        """Called when a remove button is clicked."""
        self.remove_effect(instance_id)

    def remove_effect(self, instance_id):
        """Remove a tracked effect."""
        self._active_effects.pop(instance_id, None)
        self._effect_bars.pop(instance_id, None)
        self._effect_hboxes.pop(instance_id, None)
        self._effect_remove_btns.pop(instance_id, None)
        self.redraw_effects()

    def on_parser_event(self, event):
        """Called when a parser event is triggered."""
        _LOGGER.info("Detected event %s", event)
        if isinstance(event, SpellEventCast):
            if event.duration is None:
                return
            self._active_effects[event.instance_id] = event

            bar = Gtk.ProgressBar()
            bar.set_show_text(True)
            self._effect_bars[event.instance_id] = bar

            remove_btn = Gtk.Button(image=REMOVE_IMAGE)
            self._effect_remove_btns[event.instance_id] = remove_btn
            remove_btn.connect(
                "clicked", lambda *_: self.on_remove_click(event.instance_id))
            remove_btn.get_style_context().add_class("remove")

            hbox = Gtk.HBox()
            hbox.pack_start(bar, True, True, 5)
            hbox.pack_start(remove_btn, False, False, 5)
            self._effect_hboxes[event.instance_id] = hbox

            self.redraw_effects()

        elif isinstance(event, SpellEventExpired):
            self.remove_effect(event.instance_id)

    def on_exit(self):
        """Called when the application is closing."""
        self.parser.stop()



"""User interface module for symbol_please."""
import logging

import gi

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk  # noqa: E402

_LOGGER = logging.getLogger(__name__)


class ProfileEditWindow(Gtk.Dialog):
    """Main overlay window."""

    def __init__(self, parent, profile=None):
        """Initialize the main window."""
        super().__init__("Symbol Please", parent, 0, (
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
            Gtk.STOCK_OK, Gtk.ResponseType.OK))

        self.response = Gtk.ResponseType.CANCEL

        self.set_type_hint(Gdk.WindowTypeHint.UTILITY)

        vbox = Gtk.VBox(spacing=5)
        vbox.set_spacing(5)

        label = Gtk.Label("Profile Name:")
        label.set_xalign(0)
        vbox.pack_start(label, False, False, 5)

        self.name_entry = Gtk.Entry()
        vbox.pack_start(self.name_entry, False, False, 5)

        label = Gtk.Label("Log file:")
        label.set_xalign(0)
        vbox.pack_start(label, False, False, 5)

        logfilehbox = Gtk.HBox(spacing=5)

        self.logfile_entry = Gtk.Entry()
        self.logfile_entry.set_size_request(400, 0)

        logfilehbox.pack_start(self.logfile_entry, True, True, 5)

        open_image = Gtk.Image(stock=Gtk.STOCK_OPEN)
        logfile_open_button = Gtk.Button(image=open_image)
        logfile_open_button.connect("clicked", self.on_logfile_open_clicked)

        logfilehbox.pack_start(logfile_open_button, False, False, 5)

        vbox.pack_start(logfilehbox, False, False, 5)

        label = Gtk.Label("Character Level:")
        label.set_xalign(0)
        vbox.pack_start(label, False, False, 5)

        self.level_entry = Gtk.SpinButton.new_with_range(1, 60, 1)
        vbox.pack_start(self.level_entry, False, False, 5)

        self.get_content_area().add(vbox)
        self.show_all()

    def on_logfile_open_clicked(self, widget):
        """Select the log file."""
        dialog = Gtk.FileChooserDialog(
            "Please select the log file for this profile", self,
            Gtk.FileChooserAction.OPEN, (
                Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                Gtk.STOCK_OPEN, Gtk.ResponseType.OK))

        filter_logs = Gtk.FileFilter()
        filter_logs.set_name("Log files (*.txt)")
        filter_logs.add_pattern("*.txt")
        dialog.add_filter(filter_logs)

        response = dialog.run()

        if response == Gtk.ResponseType.OK:
            self.logfile_entry.set_text(dialog.get_filename())

        dialog.destroy()

    @property
    def profile_name(self):
        """Get the entered profile name."""
        return self.name_entry.get_text()

    @profile_name.setter
    def profile_name(self, value):
        """Set the entered profile name."""
        self.name_entry.set_text(value)

    @property
    def profile_config(self):
        """Get the entered profile config."""
        log_file = self.logfile_entry.get_text()
        level = self.level_entry.get_value()

        return {
            'log_file': log_file,
            'level': level,
        }

    @profile_config.setter
    def profile_config(self, value):
        """Set the entered profile config."""
        self.logfile_entry.set_text(value.get('log_file', ''))
        self.level_entry.set_value(value.get('level', 1))



"""Tests for the parser log line."""
from datetime import datetime

import pytest

from symbol_please.parser import LogLine

def test_parse_components():
    """Test that components are split."""
    line = LogLine("[Wed Mar 20 18:40:54 2019] Your spell fizzles!", 1)

    assert line.line_id == 1
    assert line.raw_line == "[Wed Mar 20 18:40:54 2019] Your spell fizzles!"
    assert line.raw_timestamp == "Wed Mar 20 18:40:54 2019"
    assert line.timestamp == datetime(2019, 3, 20, 18, 40, 54)
    assert line.message == "Your spell fizzles!"

def test_invalid_date():
    """Test that components are split."""
    with pytest.raises(ValueError):
        line = LogLine("[Jarbled line] Your spell fizzles!", 1)

def test_invalid_line_format():
    """Test that components are split."""
    with pytest.raises(ValueError):
        line = LogLine("Wed Mar 20 18:40:54 2019 Your spell fizzles!", 1)



#!/usr/bin/env python
import json
import os
import shutil
import sys
import time
import urllib.request
import urllib.parse

import click

ENDPOINT = 'https://wiki.project1999.com/{}?action=raw'

# Get the /data/raw/spell directory
DATA_DIR = os.path.join(os.path.abspath(os.path.join(os.path.dirname(
    os.path.realpath(__file__)), os.pardir)), 'data', 'raw', 'spell')
SPELL_FILE = os.path.join(os.path.abspath(os.path.join(os.path.dirname(
    os.path.realpath(__file__)), os.pardir)), 'data', 'raw', 'all_spells.json')

NEEDS_DISAMBIGUATION = {
    "Sacrifice": "Sacrifice_(spell)",
    "Shield of Thorns": "Shield_of_Thorns_(Spell)",
    "Tishan's Discord": "Tishan`s_Discord",
    "Zumaik's Animation": "Zumaik`s_Animation",
}

if not click.confirm(
        "Redownload wiki data from Project1999 servers?"
        " This isn't necessary unless the locally committed data"
        " is out of date.", default=False):
    sys.exit(1)

with open(SPELL_FILE, 'r') as spell_file:
    all_spells = json.load(spell_file)

for spell_name in all_spells:
    if spell_name in NEEDS_DISAMBIGUATION.keys():
        # Needs to be disambiguated
        spell_name_url = NEEDS_DISAMBIGUATION[spell_name]
    else:
        spell_name_url = spell_name.replace(" ", "_")

    url = ENDPOINT.format(urllib.parse.quote_plus(spell_name_url))
    file_name = os.path.join(DATA_DIR, spell_name.replace(
        "'", "").replace(":", ""))
    print("Downloading {} from {} to file {}".format(
        spell_name, url, file_name))
    # Download the file from `url` and save it locally under `file_name`:
    with urllib.request.urlopen(url) as resp, open(file_name, 'wb') as out:
        shutil.copyfileobj(resp, out)

    # Spread out our requests to the server so we're no hammering it.
    time.sleep(1)



"""Console script for symbol_please."""
import logging
import os
import sys

import click

import symbol_please
from symbol_please.config import load_config
import symbol_please.ui


_LOGGER = logging.getLogger(__name__)


def setup_logging():
    """Set up logging format.

    Copped from home assistant:
    https://github.com/home-assistant/home-assistant/blob/dc186841e2f387efecd3746c755ccb6b8ed51b0b/homeassistant/bootstrap.py#L226-L265
    """
    fmt = ("%(asctime)s %(levelname)s (%(threadName)s) "
           "[%(name)s] %(message)s")
    datefmt = '%Y-%m-%d %H:%M:%S'

    loglevel = os.environ.get("LOGLEVEL", "INFO")

    try:
        from colorlog import ColoredFormatter
        # basicConfig must be called after importing colorlog in order to
        # ensure that the handlers it sets up wraps the correct streams.
        logging.basicConfig(level=loglevel)

        colorfmt = "%(log_color)s{}%(reset)s".format(fmt)
        logging.getLogger().handlers[0].setFormatter(ColoredFormatter(
            colorfmt,
            datefmt=datefmt,
            reset=True,
            log_colors={
                'DEBUG': 'cyan',
                'INFO': 'green',
                'WARNING': 'yellow',
                'ERROR': 'red',
                'CRITICAL': 'red',
            }
        ))
    except ImportError:
        pass

    # If the above initialization failed for any reason, setup the default
    # formatting.  If the above succeeds, this wil result in a no-op.
    logging.basicConfig(format=fmt, datefmt=datefmt, level=loglevel)


@click.command()
def main(args=None):
    """Console script for symbol_please."""
    setup_logging()
    config = load_config()
    return symbol_please.ui.main(config)


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



#!/usr/bin/env python
import json
import os
import re

import wikitextparser as wtp

# Get the /data/raw/spell directory
DATA_DIR = os.path.join(os.path.abspath(os.path.join(os.path.dirname(
    os.path.realpath(__file__)), os.pardir)), 'data', 'raw', 'spell')
SPELL_FILE = os.path.join(os.path.abspath(os.path.join(os.path.dirname(
    os.path.realpath(__file__)), os.pardir)), 'data', 'raw', 'all_spells.json')
DATA_FILE = os.path.join(os.path.abspath(os.path.join(os.path.dirname(
    os.path.realpath(__file__)), os.pardir)),
    'symbol_please', 'data', 'spell_data.json')

LEVEL_REGEX = re.compile(r' @L(?P<level>\d+)', re.IGNORECASE)
SECONDS_REGEX = re.compile(r'(?P<time>\d+(\.\d+)?) second(s)?', re.IGNORECASE)
TICKS_REGEX = re.compile(r'(?P<time>\d+(\.\d+)?) tick(s)?', re.IGNORECASE)
MINUTES_REGEX = re.compile(r'(?P<time>\d+(\.\d+)?) minute(s)?', re.IGNORECASE)
HOURS_REGEX = re.compile(r'(?P<time>\d+(\.\d+)?) hour(s)?', re.IGNORECASE)

spell_data = {}

# The wiki uses the incorrect names
NAME_FIXES = {
    'Improved Invis to Undead': 'Improved Invis vs Undead',
    'Unswerving Hammer': 'Unswerving Hammer of Faith',
}


def parse_duration_string(duration):
    """Parse a duration string and return seconds, level."""
    seconds = 0
    level = None

    seconds_match = SECONDS_REGEX.search(duration)
    if seconds_match is not None:
        seconds += float(seconds_match.group('time'))

    ticks_match = TICKS_REGEX.search(duration)
    if ticks_match is not None:
        seconds += float(ticks_match.group('time')) * 6

    minutes_match = MINUTES_REGEX.search(duration)
    if minutes_match is not None:
        seconds += float(minutes_match.group('time')) * 60

    hours_match = HOURS_REGEX.search(duration)
    if hours_match is not None:
        seconds += float(hours_match.group('time')) * 3600

    seconds = int(seconds)

    level_match = LEVEL_REGEX.search(duration)
    if level_match is not None:
        level = int(level_match.group('level'))

    return seconds, level


with open(SPELL_FILE, 'r') as spell_file:
    all_spells = json.load(spell_file)

for spell_name in all_spells:
    file_name = os.path.join(DATA_DIR, spell_name.replace(
        "'", "").replace(":", ""))

    spell_name = NAME_FIXES.get(spell_name, spell_name)
    print("Parsing {}".format(spell_name))
    with open(file_name, 'r') as class_file:
        parsed = wtp.parse(class_file.read())

    spellpage = next(
        t for t in parsed.templates if t.normal_name() in (
            'Spellpage', 'Spellpagesmart'))

    duration = spellpage.get_arg('duration').value.strip()
    target_type = spellpage.get_arg('target_type').value.strip()
    msg_cast_on_you = spellpage.get_arg('msg_cast_on_you').value.strip()
    msg_cast_on_other = spellpage.get_arg('msg_cast_on_other').value.strip()
    msg_wears_off = spellpage.get_arg('msg_wears_off').value.strip()

    # Clean up data
    if "Instant" in duration or "Up to" in duration:
        duration = None
    elif ' to ' in duration:
        duration_parts = duration.split(" to ")
        start_seconds, start_level = parse_duration_string(duration_parts[0])
        end_seconds, end_level = parse_duration_string(duration_parts[1])
        if start_level is None or end_level is None:
            raise ValueError(
                "Error getting level from multipart duration '{}'".format(
                    duration))
        duration = {
            "start": {
                "level": start_level,
                "duration": start_seconds,
            },
            "end": {
                "level": end_level,
                "duration": end_seconds,
            },
        }
    else:
        duration, _ = parse_duration_string(duration)

    if "Group" in target_type:
        target_type = "Group"
    elif target_type in (
            "Line of Sight", "Pet", "Animal", "Summoned", "Undead", "Plant",
            "Corpse", "Lifetap", "Uber Giants", "Uber Dragons"):
        target_type = "Single"
    elif target_type in ("Targeted AE", "PB AE"):
        target_type = "AE"
    elif target_type not in ("Single", "Self"):
        raise ValueError("Unknown target type '{}'".format(target_type))

    msg_cast_on_other = msg_cast_on_other.replace(
        "Someone  ", "Someone ").replace("Someone 's", "Someone's")

    spell_data[spell_name] = {
        'name': spell_name,
        'duration': duration,
        'target_type': target_type,
        'msg_cast_on_you': msg_cast_on_you,
        'msg_cast_on_other': msg_cast_on_other,
        'msg_wears_off': msg_wears_off,
    }

with open(DATA_FILE, 'w') as data_file:
    json.dump(spell_data, data_file, indent=2)



"""Spell events module."""
import re

LEVEL_UP_REGEX = re.compile(
    r'^You have gained a level! Welcome to level (?P<level>\d+)!$')


class NewLevelEvent():
    """A new level event."""
    __slots__ = ('level')

    @classmethod
    def create_events(cls, client, line):
        """Parses every log line, and returns new events to be tracked."""
        level_up_match = LEVEL_UP_REGEX.match(line.message)
        if level_up_match is not None:
            level = int(level_up_match.group('level'))
            level = max(1, min(60, level))
            client.send_event(NewLevelEvent(level))

    def __init__(self, level):
        """Create a new spell effect."""
        self.level = level



"""Parsing client for symbol_please."""
import asyncio
import logging
import queue

from symbol_please.config import update_config
from symbol_please.parser.level import NewLevelEvent
from symbol_please.parser.line import LogLine
from symbol_please.parser.spell import SpellEventHandler

_LOGGER = logging.getLogger(__name__)


class ParserClient():
    """The client that reads log lines and manages state."""

    def __init__(self, profile_name, config):
        """Create the parser client."""
        self.profile_name = profile_name
        self.config = config
        self.level = int(self.config['level'])
        self._event_listeners_thread = set()
        self._last_line_id = 0
        self._event_handlers = set()

    def add_event_listener_thread(self):
        """Returns a thread-safe queue which will emit log events."""
        listener = queue.Queue()
        self._event_listeners_thread.add(listener)
        return listener

    def remove_event_listener_thread(self, listener):
        """Deregister a listener thread if it is no longer needed."""
        self._event_listeners_thread.remove(listener)

    def parse_raw_line(self, raw_line):
        """Parse a raw line of input."""
        self._last_line_id += 1
        line_id = self._last_line_id
        try:
            line = LogLine(raw_line, line_id)
        except ValueError as ex:
            _LOGGER.warning("Unable to parse log line '%s'", line, exc_info=ex)
            return

        # Core handlers
        NewLevelEvent.create_events(self, line)

        _LOGGER.debug("Parsing log line: '%s'", line)
        for handler in self._event_handlers:
            if handler.line_queue is None:
                continue
            handler.line_queue.put_nowait(line)

        new_handlers = SpellEventHandler.create_events(self, line)
        if new_handlers:
            self._event_handlers = self._event_handlers.union(new_handlers)
            for handler in new_handlers:
                asyncio.ensure_future(self.run_handler(handler))

    async def run_handler(self, handler):
        """Run the given handler, and deregister when complete."""
        try:
            await handler.run()
        finally:
            self._event_handlers.remove(handler)

    def send_event(self, event):
        """Send an event to registered listeners."""
        _LOGGER.info("Sending event %s", event)

        # Core events
        if isinstance(event, NewLevelEvent):
            self.level = event.level
            self.config['level'] = event.level
            update_config(self.profile_name, self.config)

        for handler in self._event_handlers:
            if handler.event_queue is None:
                continue
            handler.event_queue.put_nowait(event)

        for thread_listener in self._event_listeners_thread.copy():
            thread_listener.put_nowait(event)



"""Spell events module."""
import asyncio
import datetime
import logging
import re

from symbol_please.data import SPELL_DATA

_LOGGER = logging.getLogger(__name__)

CASTING_REGEX = re.compile(r'^You begin casting (?P<spell>.+)\.$')
INTERRUPTED_REGEX = re.compile(r"^Your spell is interrupted\.$")
DID_NOT_TAKE_HOLD_REGEX = re.compile(r"^Your spell did not take hold\.$")
LOADING_REGEX = re.compile(r"^LOADING, PLEASE WAIT\.\.\.$")

# We'll cache the regexes we use to wait for spell messages
SPELL_CAST_REGEXES = {}
SPELL_WEARS_OFF_REGEXES = {}


class SpellEvent():
    """A detected spell event."""
    __slots__ = ('instance_id', 'spell')

    def __init__(self, instance_id, spell):
        """Create a new spell effect."""
        self.instance_id = instance_id
        self.spell = spell


class SpellEventCasting(SpellEvent):
    """A spell is being cast."""


class SpellEventInterrupted(SpellEvent):
    """A spell cast was interrupted."""


class SpellEventDidNotTakeHold(SpellEvent):
    """A spell cast did not take hold."""


class SpellEventResisted(SpellEvent):
    """A spell cast was resisted."""


class SpellEventTargeted(SpellEvent):
    """A spell event with a target detected."""
    __slots__ = ('target',)

    def __init__(self, instance_id, spell, target):
        """Create a new spell effect."""
        super().__init__(instance_id, spell)
        self.target = target


class SpellEventCast(SpellEventTargeted):
    """A spell has been successfully cast."""
    __slots__ = ('cast_time', 'duration')

    def __init__(self, instance_id, spell, target, cast_time, duration):
        """Create a new spell effect."""
        super().__init__(instance_id, spell, target)
        self.cast_time = cast_time
        self.duration = duration


class SpellEventExpired(SpellEventTargeted):
    """A spell effect has expired."""


class LineListener():
    """A regex log listener."""

    def __init__(self, name, pattern):
        """Create the listener."""
        self.name = name
        self.pattern = pattern
        self.line = None
        self.match = None

    def check_line(self, line):
        """Check if the given line matches."""
        _LOGGER.debug("Pattern %s", self.pattern)
        self.match = self.pattern.match(line.message)
        if self.match is not None:
            self.line = line
            return True
        return False


class SpellEventHandler():
    """Manages tracking of spell events."""

    @classmethod
    def create_events(cls, client, line):
        """Parses every log line, and returns new events to be tracked."""
        spell_match = CASTING_REGEX.match(line.message)
        if spell_match is not None:
            spell_name = spell_match.group('spell')
            if spell_name not in SPELL_DATA:
                _LOGGER.warning("No data for spell '%s'", spell_name)
                return
            spell = SPELL_DATA[spell_name]

            if not spell['msg_cast_on_you'] and not spell['msg_cast_on_other']:
                _LOGGER.info("No trackable logging for spell '%s'", spell)
                return

            return set((cls(client, line, spell),))

    def __init__(self, client, line, spell):
        """Create a new spell effect handler."""
        self.client = client
        self.line = line
        self.spell = spell
        # Start tracking new log lines as soon as the handler is constructed
        self.line_queue = asyncio.Queue()
        # We don't need to watch for other spell events by default
        self.event_queue = None

    def _send_event(self, event):
        """Send an event to the parser client."""
        # Pass the event up to the client
        self.client.send_event(event)

    async def _wait_for_messages(self, messages, timeout=None):
        """Wait for possible effect messages, with a given timeout."""
        listeners = tuple(
            LineListener(name, message) for name, message in messages.items())

        return await asyncio.wait_for(
            self.wait_for_listeners(listeners), timeout)

    async def wait_for_event(self, check_event):
        """Read events until a match is found."""
        if self.event_queue is None:
            self.event_queue = asyncio.Queue()
        while True:
            event = await self.event_queue.get()
            if check_event(event):
                return event

    async def wait_for_listeners(self, listeners):
        """Parse a line of input."""
        while True:
            line = await self.line_queue.get()
            for listener in listeners:
                if listener.check_line(line):
                    return listener

    async def run(self):
        """Handle the casting of a spell"""
        instance_id = self.line.line_id
        self._send_event(SpellEventCasting(instance_id, self.spell))

        if self.spell['name'] not in SPELL_CAST_REGEXES:
            # Always need to listen for interruped messages
            regexes = {
                'interrupted': INTERRUPTED_REGEX,
                'did_not_take_hold': DID_NOT_TAKE_HOLD_REGEX,
                'loading': LOADING_REGEX,
                'resisted': re.compile("^{}$".format(
                    re.escape("Your target resisted the {} spell.".format(
                        self.spell['name'])))),
            }
            if self.spell['msg_cast_on_you']:
                # Listen for the spell to effect self
                regexes['self'] = re.compile("^{}$".format(
                    re.escape(self.spell['msg_cast_on_you'])))
            if self.spell['msg_cast_on_other']:
                # Listen for the spell to effect others
                regexes['other'] = re.compile("^{}$".format(
                    re.escape(self.spell['msg_cast_on_other']).replace(
                        "Someone", r"(?P<target>.+)")))
            SPELL_CAST_REGEXES[self.spell['name']] = regexes

            # Listen for the spell to wear off
            if self.spell['msg_wears_off']:
                SPELL_WEARS_OFF_REGEXES[self.spell['name']] = {
                    'wears_off': re.compile("^{}$".format(
                        re.escape(self.spell['msg_wears_off'])))
                }
        try:
            result = await self._wait_for_messages(
                SPELL_CAST_REGEXES[self.spell['name']], 30)
        except asyncio.TimeoutError as ex:
            _LOGGER.warning("No cast result message detected for spell '%s'",
                            self.spell['name'], exc_info=ex)
            return
        if self.spell['target_type'] == 'Group':
            # Skip spell did not take hold messages if casting group spell to
            # see if it took effect on anyone
            while result.name == 'did_not_take_hold':
                try:
                    result = await self._wait_for_messages(
                        SPELL_CAST_REGEXES[self.spell['name']], 1)
                except asyncio.TimeoutError:
                    break

        if result.name == 'interrupted':
            self._send_event(SpellEventInterrupted(instance_id, self.spell))
            return

        if result.name == 'loading':
            self._send_event(SpellEventInterrupted(instance_id, self.spell))
            return

        if result.name == 'did_not_take_hold':
            self._send_event(SpellEventDidNotTakeHold(instance_id, self.spell))
            return

        if result.name == 'resisted':
            self._send_event(SpellEventResisted(instance_id, self.spell))
            return

        if self.spell['target_type'] == 'Group':
            target = 'GROUP'
        elif result.name == 'self':
            target = 'YOU'
        else:
            target = result.match.group('target')

        cast_time = datetime.datetime.now()
        duration = self._calculate_duration()

        self._send_event(SpellEventCast(
            instance_id, self.spell, target, cast_time, duration))

        if duration is None:
            return

        _LOGGER.info("Spell in effect: '%s'", self.spell['name'])

        def check_event(event):
            """Callback to check for spell replacement event."""
            if not isinstance(event, SpellEventCast):
                return False
            if event.spell['name'] != self.spell['name']:
                return False
            if event.instance_id == instance_id:
                return False
            if event.target != target:
                return False
            return True

        wait_for_event = self.wait_for_event(check_event)
        if target == 'YOU' and self.spell['name'] in SPELL_WEARS_OFF_REGEXES:
            wait_for_messages = self._wait_for_messages(
                SPELL_WEARS_OFF_REGEXES[self.spell['name']])
            done, pending = await asyncio.wait(
                (wait_for_messages, wait_for_event),
                timeout=duration, return_when=asyncio.FIRST_COMPLETED)
            for task in pending:
                # Cancel tasks that didn't finish first
                task.cancel()
        else:
            try:
                await asyncio.wait_for(wait_for_event, duration)
            except asyncio.TimeoutError:
                pass
        _LOGGER.info("Spell expired: '%s'", self.spell['name'])

        self._send_event(SpellEventExpired(instance_id, self.spell, target))

    def _calculate_duration(self):
        """Calculate the duration for a spell cast."""
        duration = self.spell['duration']
        if not isinstance(duration, dict):
            return duration

        _LOGGER.debug("Level %s", self.client.level)

        minlevel = duration['start']['level']
        maxlevel = duration['end']['level']
        minduration = duration['start']['duration']
        maxduration = duration['end']['duration']
        if self.client.level <= minlevel:
            return minduration
        if self.client.level >= maxlevel:
            return maxduration

        # Interpolate
        fraction = (self.client.level - minlevel) / (maxlevel - minlevel)
        return ((maxduration - minduration) * fraction) + minduration



#!/usr/bin/env python
import os
import shutil
import sys
import time
import urllib.request

import click

ENDPOINT = 'https://wiki.project1999.com/{}?action=raw'
CLASSES = [
    'Enchanter',
    'Magician',
    'Necromancer',
    'Wizard',
    'Cleric',
    'Druid',
    'Shaman',
    'Bard',
    'Monk',
    'Ranger',
    'Rogue',
    'Paladin',
    'Shadow_Knight',
    'Warrior',
]

# Get the /data/raw/class directory
DATA_DIR = os.path.join(os.path.abspath(os.path.join(os.path.dirname(
    os.path.realpath(__file__)), os.pardir)), 'data', 'raw', 'class')

if not click.confirm(
        "Redownload wiki data from Project1999 servers?"
        " This isn't necessary unless the locally committed data"
        " is out of date.", default=False):
    sys.exit(1)

for class_name in CLASSES:
    url = ENDPOINT.format(class_name)
    file_name = os.path.join(DATA_DIR, class_name)
    print("Downloading {} from {} to file {}".format(
        class_name, url, file_name))
    # Download the file from `url` and save it locally under `file_name`:
    with urllib.request.urlopen(url) as resp, open(file_name, 'wb') as out:
        shutil.copyfileobj(resp, out)

    # Spread out our requests to the server so we're no hammering it.
    time.sleep(1)



"""User interface module for symbol_please."""
import os

import gi

from symbol_please.ui.profile_select_window import ProfileSelectWindow

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk  # noqa: E402

STYLESHEET = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), 'res', 'style.css')


def main(config):
    """Initialize the UI."""
    style_provider = Gtk.CssProvider()
    with open(STYLESHEET, 'rb') as f:
        style_provider.load_from_data(f.read())

    Gtk.StyleContext.add_provider_for_screen(
        Gdk.Screen.get_default(), style_provider,
        Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
    )

    win = ProfileSelectWindow(config)
    win.show_all()

    try:
        Gtk.main()
    finally:
        win.on_exit()

    return 0



#!/usr/bin/env python
import json
import os

import wikitextparser as wtp

CLASSES = [
    'Enchanter',
    'Magician',
    'Necromancer',
    'Wizard',
    'Cleric',
    'Druid',
    'Shaman',
    'Bard',
    'Monk',
    'Ranger',
    'Rogue',
    'Paladin',
    'Shadow_Knight',
    'Warrior',
]

# Get the /data/raw/class directory
DATA_DIR = os.path.join(os.path.abspath(os.path.join(os.path.dirname(
    os.path.realpath(__file__)), os.pardir)), 'data', 'raw', 'class')
SPELL_FILE = os.path.join(os.path.abspath(os.path.join(os.path.dirname(
    os.path.realpath(__file__)), os.pardir)), 'data', 'raw', 'all_spells.json')

all_spells = set()

for class_name in CLASSES:
    file_name = os.path.join(DATA_DIR, class_name)
    print("Parsing {}".format(class_name))
    with open(file_name, 'r') as class_file:
        parsed = wtp.parse(class_file.read())
    for template in parsed.templates:
        if template.normal_name() != 'SpellRow':
            continue
        name_arg = template.get_arg('name')
        all_spells.add(name_arg.value.strip())

sorted_spells = sorted(all_spells)
with open(SPELL_FILE, 'w') as spell_file:
    json.dump(sorted_spells, spell_file, indent=2)



"""User interface module for symbol_please."""
import logging

import gi

from symbol_please.config import save_config
from symbol_please.ui.overlay_window import OverlayWindow
from symbol_please.ui.profile_edit_window import ProfileEditWindow

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk  # noqa: E402

_LOGGER = logging.getLogger(__name__)

REMOVE_IMAGE = Gtk.Image(stock=Gtk.STOCK_REMOVE)


class ProfileSelectWindow(Gtk.Window):
    """Main overlay window."""

    def __init__(self, config):
        """Initialize the main window."""
        super().__init__(title="Symbol Please")

        self.overlay_window = None

        self.config = config

        self.connect("destroy", self.on_destroy)
        self.set_type_hint(Gdk.WindowTypeHint.UTILITY)

        vbox = Gtk.VBox(spacing=5)
        vbox.set_spacing(5)

        menuhbox = Gtk.HBox()

        self.new_button = Gtk.Button(label="New")
        self.new_button.connect("clicked", self.on_new_clicked)
        menuhbox.pack_start(self.new_button, False, False, 5)

        self.edit_button = Gtk.Button(label="Edit")
        self.edit_button.connect("clicked", self.on_edit_clicked)
        menuhbox.pack_start(self.edit_button, False, False, 5)

        self.delete_button = Gtk.Button(label="Delete")
        self.delete_button.connect("clicked", self.on_delete_clicked)
        menuhbox.pack_start(self.delete_button, False, False, 5)

        vbox.pack_start(menuhbox, False, False, 5)

        self.select_listbox = Gtk.ListBox()
        self.select_listbox.set_size_request(400, 400)
        self.select_listbox.set_activate_on_single_click(False)
        self.select_listbox.connect("row-selected", self.on_row_selected)
        self.select_listbox.connect("row-activated", self.on_row_activated)

        vbox.pack_start(self.select_listbox, True, True, 5)

        starthbox = Gtk.HBox()

        self.start_button = Gtk.Button(label="Start")
        self.start_button.connect("clicked", self.on_start_clicked)
        starthbox.pack_end(self.start_button, False, False, 5)

        vbox.pack_start(starthbox, False, False, 5)

        self.add(vbox)

        self.update_profiles()
        self.update_buttons()

    def update_profiles(self):
        """Update the profile list."""
        for child in self.select_listbox.get_children():
            child.destroy()

        for profile in self.config['profiles'].keys():
            label = Gtk.Label(profile, xalign=0)
            self.select_listbox.add(label)
            label.show()

    @property
    def selected_profile_name(self):
        """Get the selected profile name."""
        selected_row = self.select_listbox.get_selected_row()
        if selected_row is None:
            return None
        return selected_row.get_children()[0].get_text()

    @property
    def selected_profile_config(self):
        """Get the selected profile name."""
        name = self.selected_profile_name
        if name is None:
            return None
        return self.config['profiles'].get(name)

    def update_buttons(self):
        """Update the enabled status of buttons."""
        enabled = self.select_listbox.get_selected_row() is not None
        self.edit_button.set_sensitive(enabled)
        self.delete_button.set_sensitive(enabled)
        self.start_button.set_sensitive(enabled)

    def on_row_selected(self, widget, row):
        """Called when the selected row changes."""
        self.update_buttons()

    def on_row_activated(self, widget, row):
        """Called when the selected row is activated."""
        self.start()

    def on_new_clicked(self, widget):
        """Create a new profile."""
        edit_window = ProfileEditWindow(self)
        response = edit_window.run()
        if response == Gtk.ResponseType.OK:
            self.config['profiles'][edit_window.profile_name] = \
                edit_window.profile_config
            save_config(self.config)
            self.update_profiles()

        edit_window.destroy()

    def on_edit_clicked(self, widget):
        """Edit a profile."""
        edit_window = ProfileEditWindow(self)
        edit_window.profile_name = self.selected_profile_name
        edit_window.profile_config = self.selected_profile_config

        response = edit_window.run()
        if response == Gtk.ResponseType.OK:
            self.config['profiles'][edit_window.profile_name] = \
                edit_window.profile_config
            save_config(self.config)
            self.update_profiles()

        edit_window.destroy()

    def on_delete_clicked(self, widget):
        """Delete a profile."""
        del self.config['profiles'][self.selected_profile_name]
        save_config(self.config)
        self.update_profiles()

    def on_start_clicked(self, widget):
        """Start parsing for the selected profile."""
        self.start()

    def on_destroy(self, widget):
        """Window is destroyed."""
        if not self.overlay_window:
            Gtk.main_quit()

    def start(self):
        """Start parsing the selected profile."""
        self.overlay_window = OverlayWindow(
            self.selected_profile_name, self.selected_profile_config)
        self.overlay_window.show_all()
        self.destroy()

    def on_exit(self):
        """Called when the application is closing."""
        if self.overlay_window is not None:
            self.overlay_window.on_exit()



"""The setup script."""

from setuptools import setup, find_packages

import symbol_please

with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = [
    'Click>=6.0',
    'colorlog>=4.0.2',
    'python-dateutil>=2.8.0',
]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', ]

setup(
    author=symbol_please.__author__,
    author_email=symbol_please.__email__,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: X11 Applications :: GTK',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
    ],
    description="Symbol Please is a visual log parsing helper for Project 1999",
    entry_points={
        'console_scripts': [
            'symbol_please=symbol_please.cli:main',
        ],
    },
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme,
    include_package_data=True,
    keywords='symbol_please',
    name='symbol_please',
    packages=find_packages(include=['symbol_please']),
    package_data={'symbol_please': ['data/spell_data.json']},
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://gitlab.com/armills/symbol_please',
    version=symbol_please.__version__,
    zip_safe=False,
)



"""Console script for symbol_please."""
import logging
import json
import os

import symbol_please

_LOGGER = logging.getLogger(__name__)

CONFIG_HOME = os.environ.get(
    "XDG_CONFIG_HOME", os.path.join(os.path.expanduser('~'), '.config'))
CONFIG_FILE_DIR = os.path.join(CONFIG_HOME, symbol_please.__name__)
CONFIG_FILE_PATH = os.path.join(CONFIG_FILE_DIR, 'settings.json')

VERSION = 1


def load_config():
    """Load configuration file."""
    if not os.path.isfile(CONFIG_FILE_PATH):
        return {
            '_version': VERSION,
            'profiles': {},
        }

    _LOGGER.info("Loading config file '%s'", CONFIG_FILE_PATH)
    with open(CONFIG_FILE_PATH, 'r') as config_file:
        config = json.load(config_file)

        if config.get('_version') != VERSION:
            raise ValueError("Invalid config version detected")

        return config


def save_config(config):
    """Save the configuration file."""
    os.makedirs(CONFIG_FILE_DIR, exist_ok=True)

    _LOGGER.info("Saving config file '%s'", CONFIG_FILE_PATH)
    with open(CONFIG_FILE_PATH, 'w') as config_file:
        return json.dump(config, config_file, indent=2)


def update_config(profile_name, profile_config):
    """Update the given profile."""
    config = load_config()
    config['profiles'][profile_name] = profile_config
    save_config(config)



"""Parsing thread for symbol_please."""
import asyncio
import logging
import threading

from symbol_please.parser.client import ParserClient

_LOGGER = logging.getLogger(__name__)


class ParserThread(threading.Thread):
    """Log parsing thread."""

    def __init__(self, profile_name, config):
        """Create the parser thread object."""
        super().__init__(name=self.__class__.__name__)
        self._config = config
        self._exit_event = threading.Event()
        self._client = ParserClient(profile_name, config)

    def run(self):
        """Runs the log parsing thread."""
        filename = self._config['log_file']
        _LOGGER.info("Opening log file '%s'", filename)
        with open(filename, buffering=1) as logfile:
            # Read to current file position
            while logfile.readline():
                continue

            # Run the event loop
            asyncio.new_event_loop().run_until_complete(self._main(logfile))
        _LOGGER.info("Parsing thread exiting")

    async def _main(self, logfile):
        """Main coroutine function for event loop."""
        while not self._exit_event.is_set():
            # readline is technically blocking IO, but we'll allow it since
            # it'll never be more than one line, and is preferable to the
            # overhead of offloading file reading to another separate thread.
            line = logfile.readline()

            if not line:
                await asyncio.sleep(0.2)
                continue

            line = line.strip()
            if not line:
                # Skip empty lines that sneak in
                continue
            self._client.parse_raw_line(line)

    def stop(self):
        """Stop the parser thread."""
        self._exit_event.set()
        self.join()

    @property
    def client(self):
        """Get the client object."""
        return self._client
